/* auditd-plugin-lib.c --
 * Copyright 2005 Linuon Inc. Tokyo Japan
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors:
 *   Junji Kanemaru <junji.kanemaru@linuon.com>
 * 
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <sys/stat.h>
#include <dlfcn.h>

#include <libaudit.h>
#include <auditd.h>
#include "auditd-config.h"
#include "auditd-event.h"
#include "auditd-plugin-lib.h"

#undef SLEEP_THREAD_WHEN_CHACHE_LIMIT_ECCEEDED

#define SCHED_POLICY    SCHED_FIFO
#define MIN_PRIORITY    sched_get_priority_min(SCHED_POLICY)
#define MAX_PRIORITY    sched_get_priority_max(SCHED_POLICY)

#define THREAD_MAIN_LOOP_START  \
    while (!has_fault_occurred()) { \
        pthread_testcancel();

#define THREAD_MAIN_LOOP_END    }

/* Local Data */
static int all_threads_died = 0;
static plugin_api_list_t plugin_api_list;

/* Local function prototypes
*/
/* load_and_get_entry: */
static plugin_entry load_and_get_entry(const auditd_env_t* env, 
    const char* plugin);
/* register_receiver: */
static audit_event_list_t* register_receiver(
    const auditd_env_t* env,
    receiver_list_t* receiver_list,
    receive_event_method_t method,
    on_error_method_t on_error_method);
/* register_interpreter: */
static audit_event_list_t* register_interpreter(
    const auditd_env_t* env,
    audit_event_list_t* raw_event_list,
    interpreter_list_t* interpreter_list,
    interpret_event_method_t method,
    on_error_method_t on_error_method);
/* register_consumer: */
static audit_event_list_t* register_consumer(
    const auditd_env_t* env,
    consumer_list_t* consumer_list,
    audit_event_list_t* interpreted_event_list,
    consume_event_method_t method,
    on_error_method_t on_error_method);
/* Threads for plugin API's */
static void* receiver_thread(void* arg);
static void* interpreter_thread(void* arg);
static void* consumer_thread(void* arg);
/* Thread cancel handlers */
static void receiver_thread_cancel_handler(void* arg);
static void interpreter_thread_cancel_handler(void* arg);
static void consumer_thread_cancel_handler(void* arg);

/* load_and_get_entry:
** Loads plugin and get entry from it. */
static plugin_entry load_and_get_entry(
    const auditd_env_t* env, 
    const char* plugin)
{
    void* plugin_ptr;
    char* fname_tmp;
    struct stat st;
    plugin_entry pent;
    int rc;

    plugin_ptr = NULL;

    fname_tmp = (char*)malloc(strlen(plugin) + 8);
    AUDITD_ASSERT(fname_tmp != NULL);

    /* Create file name for real world */
    sprintf(fname_tmp, "plugins/%s", plugin);

    /* Check if file exists */
    if ((rc = stat(fname_tmp, &st)) < 0 || ((st.st_mode & S_IFREG) != S_IFREG)) {
        msg(LOG_ERR, "File %s not found or not a regular file (%s)", 
                fname_tmp, strerror(errno));
        free(fname_tmp);
        return NULL;
    }

    /* Open the plugin .so file */
    plugin_ptr = dlopen(fname_tmp, RTLD_NOW);
    if (plugin_ptr == NULL) {
        msg(LOG_ERR, "Loading plugin %s failed (%s)", fname_tmp, dlerror());
        free(fname_tmp);
        return NULL;
    }

    /* We don't need fname_tmp anymore */
    free(fname_tmp);

    /* We loaded something but not sure if it is valid auditd plugin yet */
    if (plugin_ptr != NULL) {
        char* error;
        auditd_plugin_type_t type;

        pent = (plugin_entry)dlsym(plugin_ptr, AUDITD_PLUGIN_ENTRY);
        /* Check error by dlerror(). If you are unsure why this then
        ** read man dlsym. error != NULL means error occurred. */
        if ((error = dlerror()) != NULL) {
            msg(LOG_ERR, "Loading symbol from plugin %s failed (%s)",
                    plugin, error);
            dlclose(plugin_ptr);
            return NULL;
        }

    }

    return pent;
}


/* register_receiver:
*/
static audit_event_list_t* register_receiver(
    const auditd_env_t* env,
    receiver_list_t* receiver_list,
    receive_event_method_t method,
    on_error_method_t on_error_method)
{
    receiver_list_t* list;
    receiver_list_t* last;

    AUDITD_ASSERT(method != NULL);

    /* Search list to find method is already registered */
    list = last = receiver_list;
    while(list != NULL) {
#if 0
        if (list->receive_event == method) {
            return list->raw_event_list;
        }
#endif
        last = list;
        list = list->next;
    }

    /* We are at head but never initialized. */
    if (last->receive_event == NULL) {
        list = last;
    }
    else {
        list = (receiver_list_t*)malloc(sizeof(receiver_list_t));
    }
    AUDITD_ASSERT(list != NULL);
    list->receive_event = method;
    list->error_handler.on_error_callback = on_error_method;
    list->error_handler.done = AUDITD_PLUGIN_THREAD_NOT_FINISHED;
    list->next = NULL;
    list->pps.stat = PPS_AVAILABLE;
    list->pps.buff = NULL;
    list->raw_event_list = 
            (audit_event_list_t*)malloc(sizeof(audit_event_list_t));
    AUDITD_ASSERT(list->raw_event_list != NULL);
    pthread_mutex_init(&list->raw_event_list->lock, NULL);
    pthread_cond_init(&list->raw_event_list->cond, NULL);
    list->raw_event_list->head = NULL;
    list->raw_event_list->tail = NULL;
    list->raw_event_list->size = 0;

    /* Make link */
    if (last != NULL && last != list) {
        last->next = list;
    }

    return list->raw_event_list;
}

/* register_interpreter:
*/
static audit_event_list_t* register_interpreter(
    const auditd_env_t* env,
    audit_event_list_t* raw_event_list,
    interpreter_list_t* interpreter_list,
    interpret_event_method_t method,
    on_error_method_t on_error_method)
{
    interpreter_list_t* list;
    interpreter_list_t* last;

    AUDITD_ASSERT(raw_event_list != NULL);
    AUDITD_ASSERT(method != NULL);

    /* Search list to find method is already registered */
    list = last = interpreter_list;
    while(list != NULL) {
#if 0
        if (list->interpret_event == method &&
            list->raw_event_list == raw_event_list) {
            return list->interpreted_event_list;
        }
#endif
        last = list;
        list = list->next;
    }

    /* We are at head but never initialized. */
    if (last->interpret_event == NULL) {
        list = last;
    }
    else {
        list = (interpreter_list_t*)malloc(sizeof(interpreter_list_t));
    }
    AUDITD_ASSERT(list != NULL);
    list->interpret_event = method;
    list->error_handler.on_error_callback = on_error_method;
    list->error_handler.done = AUDITD_PLUGIN_THREAD_NOT_FINISHED;
    list->raw_event_list = raw_event_list;
    list->next = NULL;
    list->pps.stat = PPS_AVAILABLE;
    list->pps.buff = NULL;
    list->interpreted_event_list = 
            (audit_event_list_t*)malloc(sizeof(audit_event_list_t));
    AUDITD_ASSERT(list->interpreted_event_list != NULL);
    pthread_mutex_init(&list->interpreted_event_list->lock, NULL);
    pthread_cond_init(&list->interpreted_event_list->cond, NULL);
    list->interpreted_event_list->head = NULL;
    list->interpreted_event_list->tail = NULL;
    list->interpreted_event_list->size = 0;

    /* Make link */
    if (last != NULL && last != list) {
        last->next = list;
    }

    return list->interpreted_event_list;
}

/* register_consumer:
*/
static audit_event_list_t* register_consumer(
    const auditd_env_t* env,
    consumer_list_t* consumer_list,
    audit_event_list_t* interpreted_event_list,
    consume_event_method_t method,
    on_error_method_t on_error_method)
{
    consumer_list_t* list;
    consumer_list_t* last;

    AUDITD_ASSERT(interpreted_event_list != NULL);
    AUDITD_ASSERT(method != NULL);

    /* Search list to find method is already registered */
    list = last = consumer_list;
    while(list != NULL) {
#if 0
        if (list->consume_event == method &&
            list->interpreted_event_list == interpreted_event_list) {
            return list->interpreted_event_list;
        }
#endif
        last = list;
        list = list->next;
    }

    /* We are at head but never initialized. */
    if (last->consume_event == NULL) {
        list = last;
    }
    else {
        list = (consumer_list_t*)malloc(sizeof(consumer_list_t));
    }
    AUDITD_ASSERT(list != NULL);
    list->consume_event = method;
    list->error_handler.on_error_callback = on_error_method;
    list->error_handler.done = AUDITD_PLUGIN_THREAD_NOT_FINISHED;
    list->next = NULL;
    list->pps.stat = PPS_AVAILABLE;
    list->pps.buff = NULL;
    list->interpreted_event_list = interpreted_event_list;

    /* Make link */
    if (last != NULL && last != list) {
        last->next = list;
    }

    return list->interpreted_event_list;
}

/* receiver_thread: 
*/
static void* receiver_thread(void* arg)
{
    const auditd_env_t* env;
    const struct daemon_conf* config;
    receiver_list_t* receiver_list;
    receive_event_method_t receive_event;
    audit_event_list_t* raw_event_list;
    auditd_plugin_type_t event_type;
    void* data;
    size_t size;
    plugin_private_storage_t* pps;
    audit_event_elem_t* elem;
    int rc;

    pthread_attr_t ta;
    struct sched_param schd_param;
    int prio_saved;
    int list_size;

    env = get_env();
    config = (const struct daemon_conf*)env->config;

    receiver_list = (receiver_list_t*)arg;
    receive_event = receiver_list->receive_event;
    raw_event_list = receiver_list->raw_event_list;
    pps = &receiver_list->pps;

    /* Increase thread priority */
    pthread_attr_init(&ta);
    pthread_attr_getschedparam(&ta, &schd_param);
    schd_param.sched_priority += config->thread_prio_inc;
    prio_saved = schd_param.sched_priority;
    pthread_attr_setschedparam(&ta, &schd_param);

    /* Set thread state cancelable */
    pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);

    /* Register pthread cancel handler */
    pthread_cleanup_push(receiver_thread_cancel_handler, 
            (void*)receiver_list);

    /* Do receive forever, never breaks unless error */
    THREAD_MAIN_LOOP_START {
        rc = (*receive_event)(&event_type, &data, &size, pps);
        if (rc == 0) {
            continue;
        }
        else if (rc < 0) {
            goto fatal_error;
        }

        elem = audit_event_add_to_tail(raw_event_list, event_type, data, size);
        if (elem == NULL) {
            AUDITD_FATAL_ERROR(ERROR_THREAD_DYING, 
                    "Receiver thread is exiting...");
            goto fatal_error;
        }

        /* Change thread priority dynamically if config->event_list_size
        ** ecceeded. However we don't wait list, list keeps growing. 
        ** TODO: Should we wait? */
        if (raw_event_list->size > config->event_list_size) {
#ifdef SLEEP_THREAD_WHEN_CHACHE_LIMIT_ECCEEDED
            struct timespec ts;
            ts.tv_sec = 0;
            ts.tv_nsec = (raw_event_list->size / config->event_list_size
                    * 5);
            nanosleep(&ts, NULL);
#endif
            if (schd_param.sched_priority != MIN_PRIORITY) {
                schd_param.sched_priority = MIN_PRIORITY;
                pthread_attr_setschedparam(&ta, &schd_param);
            }
        }
        else if (raw_event_list->size == 0) {
            if (schd_param.sched_priority != prio_saved) {
                schd_param.sched_priority = prio_saved;
                pthread_attr_setschedparam(&ta, &schd_param);
            }
        }
    } THREAD_MAIN_LOOP_END;

fatal_error:
    /* close cleanup handler brace */
    pthread_cleanup_pop(rc < 0);
}

/* interpreter_thread:
*/
static void* interpreter_thread(void* arg)
{
    const auditd_env_t* env;
    const struct daemon_conf* config;
    interpreter_list_t* interpreter_list;
    interpret_event_method_t interpret_event;
    audit_event_list_t* raw_event_list;
    audit_event_list_t* interpreted_event_list;
    auditd_plugin_type_t interpreted_type;
    void* interpreted_data;
    size_t interpreted_size;
    plugin_private_storage_t* pps;
    audit_event_elem_t* elem;
    int rc;

    pthread_attr_t ta;
    struct sched_param schd_param;
    int prio_saved;

    env = get_env();
    config = (const struct daemon_conf*)env->config;

    interpreter_list = (interpreter_list_t*)arg;
    interpret_event = interpreter_list->interpret_event;
    raw_event_list = interpreter_list->raw_event_list;
    interpreted_event_list = interpreter_list->interpreted_event_list;
    pps = &interpreter_list->pps;

    /* Increase thread priority */
    pthread_attr_init(&ta);
    pthread_attr_getschedparam(&ta, &schd_param);
    schd_param.sched_priority += config->thread_prio_inc;
    prio_saved = schd_param.sched_priority;
    pthread_attr_setschedparam(&ta, &schd_param);

    /* Set thread state cancelable */
    pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);

    /* Register pthread cancel handler */
    pthread_cleanup_push(interpreter_thread_cancel_handler, 
            (void*)interpreter_list);

    /* Never breaks unless error */
    THREAD_MAIN_LOOP_START {
        elem = audit_event_take_from_head(raw_event_list);
        if (elem == NULL) {
            continue;
        }

        rc = (*interpret_event)(elem->type, elem->data, elem->size, 
                &interpreted_type, &interpreted_data, &interpreted_size, pps);
        if (rc == 0) {
            continue;
        }
        else if (rc < 0) {
            goto fatal_error;
        }

        /* Raw data is already precceded. No longer needed */
        free(elem);

        elem = audit_event_add_to_tail(interpreted_event_list, interpreted_type,
                interpreted_data, interpreted_size);
        if (elem == NULL) {
            goto fatal_error;
        }

        /* Change thread priority dynamically if config->event_list_size
        ** ecceeded. However we don't wait list, list keeps growing. 
        ** TODO: Should we wait? */
        if (interpreted_event_list->size > config->event_list_size) {
#ifdef SLEEP_THREAD_WHEN_CHACHE_LIMIT_ECCEEDED
            struct timespec ts;
            ts.tv_sec = 0;
            ts.tv_nsec = (interpreted_event_list->size /
                    config->event_list_size * 5);
            nanosleep(&ts, NULL);
#endif
            if (schd_param.sched_priority != MIN_PRIORITY) {
                schd_param.sched_priority = MIN_PRIORITY;
                pthread_attr_setschedparam(&ta, &schd_param);
            }
        }
        else if (raw_event_list->size == 0) {
            if (schd_param.sched_priority != prio_saved) {
                schd_param.sched_priority = prio_saved;
                pthread_attr_setschedparam(&ta, &schd_param);
            }
        }
    } THREAD_MAIN_LOOP_END;

fatal_error:
    /* close cleanup handler brace */
    pthread_cleanup_pop(rc < 0);
    AUDITD_FATAL_ERROR(ERROR_THREAD_DYING, "Interpreter thread is exiting...");
}

/* consumer_thread:
*/
static void* consumer_thread(void* arg)
{
    const auditd_env_t* env;
    const struct daemon_conf* config;
    consumer_list_t* consumer_list;
    consume_event_method_t consume_event;
    audit_event_list_t* interpreted_event_list;
    plugin_private_storage_t* pps;
    audit_event_elem_t* elem;
    int rc;

    pthread_attr_t ta;
    struct sched_param schd_param;
    int prio_saved;

    env = get_env();
    config = (const struct daemon_conf*)env->config;

    consumer_list = (consumer_list_t*)arg;
    consume_event = consumer_list->consume_event;
    interpreted_event_list = consumer_list->interpreted_event_list;
    pps = &consumer_list->pps;

    /* Increase thread priority */
    pthread_attr_init(&ta);
    pthread_attr_getschedparam(&ta, &schd_param);
    schd_param.sched_priority += config->thread_prio_inc;
    prio_saved = schd_param.sched_priority;
    pthread_attr_setschedparam(&ta, &schd_param);

    /* Set thread state cancelable */
    pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);

    /* Register pthread cancel handler */
    pthread_cleanup_push(consumer_thread_cancel_handler, 
            (void*)consumer_list);

    /* Never breaks unless error */
    THREAD_MAIN_LOOP_START {
        elem = audit_event_take_from_head(interpreted_event_list);
        if (elem == NULL) {
            continue;
        }
        rc = (*consume_event)(elem->type, elem->data, elem->size, pps);
        if (rc == 0) {
            continue;
        }
        else if (rc < 0) {
            goto fatal_error;
        }
        free(elem);

        /* Change thread priority dynamically if config->event_list_size
        ** ecceeded. However we don't wait list, list keeps growing. 
        ** TODO: Should we wait? */
        if (interpreted_event_list->size > config->event_list_size) {
#ifdef SLEEP_THREAD_WHEN_CHACHE_LIMIT_ECCEEDED
            struct timespec ts;
            ts.tv_sec = 0;
            ts.tv_nsec = (interpreted_event_list->size /
                    config->event_list_size * 5);
            nanosleep(&ts, NULL);
#endif
            if (schd_param.sched_priority != MAX_PRIORITY) {
                schd_param.sched_priority = MAX_PRIORITY;
                pthread_attr_setschedparam(&ta, &schd_param);
            }
        }
        else if (interpreted_event_list->size == 0) {
            if (schd_param.sched_priority != prio_saved) {
                schd_param.sched_priority = prio_saved;
                pthread_attr_setschedparam(&ta, &schd_param);
            }
        }
    } THREAD_MAIN_LOOP_END;

fatal_error:
    /* close cleanup handler brace */
    pthread_cleanup_pop(rc < 0);
    AUDITD_FATAL_ERROR(ERROR_THREAD_DYING, "Consumer thread is exiting...");
}

/* receiver_thread_cancel_handler:
*/
static void receiver_thread_cancel_handler(void* arg)
{
    receiver_list_t* receiver_list;

    receiver_list = (receiver_list_t*)arg;

    /* We do nothig here for receivers, just kill 'em all */
    msg(LOG_ERR, "Reciver thread exitting...");
    receiver_list->error_handler.done = AUDITD_PLUGIN_THREAD_FINISHED;
    msg(LOG_ERR, "Reciver thread finished");
}

/* interpreter_thread_cancel_handler:
*/
static void interpreter_thread_cancel_handler(void* arg)
{
    interpreter_list_t* interpreter_list;
    audit_event_list_t* raw_event_list;
    audit_event_elem_t* elem;
    audit_event_elem_t* next;
    on_error_method_t error_callback;

    interpreter_list = (interpreter_list_t*)arg;

    raw_event_list = interpreter_list->raw_event_list;
    error_callback = interpreter_list->error_handler.on_error_callback;

    msg(LOG_ERR, "Interpreter thread exitting...");

    /* HACK: We don't use audit_event_take_from_head() here.
    ** It is because 1)all reciver threads are already canceled,
    ** so no one is writing to the list 2)since we are in canceling
    ** process which means this thread was in wait queue, so we
    ** can assume it was waiting on mutex in audit_event_take_from_head().
    ** See more detail in pthread_cancel() */
    elem = raw_event_list->head;
    while (elem != NULL) {
        if ((*error_callback)(elem->type, elem->data, elem->size) < 0) {
            msg(LOG_ERR, "Error handler of interpreter thread failed.");
        }
        next = elem->next;
        free(elem);
        elem = next;
    }
    raw_event_list->tail = raw_event_list->head = NULL;
    interpreter_list->error_handler.done = AUDITD_PLUGIN_THREAD_FINISHED;
    msg(LOG_ERR, "Interpreter thread finished.");
}

/* consumer_thread_cancel_handler:
*/
static void consumer_thread_cancel_handler(void* arg)
{
    consumer_list_t* consumer_list;
    audit_event_list_t* interpreted_event_list;
    audit_event_elem_t* elem;
    audit_event_elem_t* next;
    on_error_method_t error_callback;

    consumer_list = (consumer_list_t*)arg;

    interpreted_event_list = consumer_list->interpreted_event_list;
    error_callback = consumer_list->error_handler.on_error_callback;

    msg(LOG_ERR, "Consumer thread exitting...");

    /* HACK: We don't use audit_event_take_from_head() here.*/
    elem = interpreted_event_list->head;
    while (elem != NULL) {
        if ((*error_callback)(elem->type, elem->data, elem->size) < 0) {
            msg(LOG_ERR, "Error handler of consumer thread failed.");
        }
        next = elem->next;
        free(elem);
        elem = next;
    }
    interpreted_event_list->tail = interpreted_event_list->head = NULL;
    consumer_list->error_handler.done = AUDITD_PLUGIN_THREAD_FINISHED;
    msg(LOG_ERR, "Consumer thread finished.");
}


/* init_plugins:
*/
int init_plugins(
    const auditd_env_t* env, 
    const plugin_file_list_t* plugin_file_list)
{
    plugin_entry pent;
    int cnt;
    const plugin_combo_t* combo;
    const auditd_plugin_api_t* api;
    int i;

    audit_event_list_t* raw_event_list;
    audit_event_list_t* interpreted_event_list;

    cnt = plugin_file_list->cnt;
    combo = plugin_file_list->combo;

    /* Clear plugin list */
    memset(&plugin_api_list, 0, sizeof(plugin_api_list_t));

    /* Load plugins and register them */
    for (i=0; i<cnt; i++) {

        /* Load receiver from specified plugin */
        pent = load_and_get_entry(env, (const char*)combo[i].receiver_plugin);
        AUDITD_ASSERT(pent != NULL);

        api = (*pent)(env);
        AUDITD_ASSERT(api != NULL);
        AUDITD_ASSERT(api->receive_event != NULL);

        raw_event_list = register_receiver(
                env, 
                &plugin_api_list.receiver_list, 
                api->receive_event,
                api->on_error);
        AUDITD_ASSERT(raw_event_list != NULL);

        /* Load interpreter from specified plugin */
        pent = load_and_get_entry(env, combo[i].interpreter_plugin);
        AUDITD_ASSERT(pent != NULL);

        api = (*pent)(env);
        AUDITD_ASSERT(api != NULL);
        AUDITD_ASSERT(api->interpret_event != NULL);

        interpreted_event_list = register_interpreter(
                env, 
                raw_event_list,
                &plugin_api_list.interpreter_list, 
                api->interpret_event,
                api->on_error);
        AUDITD_ASSERT(interpreted_event_list != NULL);

        /* Load consumer from specified plugin */
        pent = load_and_get_entry(env, combo[i].consumer_plugin);
        AUDITD_ASSERT(pent != NULL);

        api = (*pent)(env);
        AUDITD_ASSERT(api != NULL);
        AUDITD_ASSERT(api->consume_event != NULL);

        register_consumer(
                env, 
                &plugin_api_list.consumer_list,
                interpreted_event_list, 
                api->consume_event,
                api->on_error);
    }

    return 0;
}

/* start_event_threads:
*/
int start_event_threads(
    const auditd_env_t* env)
{
    int rc;
    int cnt;
    const struct daemon_conf* config;

    receiver_list_t* receiver_list;
    interpreter_list_t* interpreter_list;
    consumer_list_t* consumer_list;
    pthread_attr_t ta;

    receiver_list = &plugin_api_list.receiver_list;
    interpreter_list = &plugin_api_list.interpreter_list;
    consumer_list = &plugin_api_list.consumer_list;

    config = (const struct daemon_conf*)env->config;

    pthread_attr_init(&ta);
    pthread_attr_setdetachstate(&ta, PTHREAD_CREATE_DETACHED);
    pthread_attr_setschedpolicy(&ta, SCHED_POLICY);

    /* start receiver threads */
    while(receiver_list != NULL && receiver_list->receive_event != NULL) {
        rc = pthread_create(&receiver_list->pthread_id, &ta, receiver_thread, 
                (void*)receiver_list);
        if (rc != 0) {
            return -1;
        }
        msg(LOG_DEBUG, "Started receiver thread(0x%08x).", 
                receiver_list->pthread_id);
        //pthread_detach(receiver_list->pthread_id);
        receiver_list = receiver_list->next;
    }

    /* start interpreter threads */
    while(interpreter_list != NULL && 
            interpreter_list->interpret_event != NULL) {
        rc = pthread_create(&interpreter_list->pthread_id, &ta, 
                interpreter_thread, (void*)interpreter_list);
        if (rc != 0) {
            return -1;
        }
        msg(LOG_DEBUG, "Started interpreter thread(0x%08x).",
                interpreter_list->pthread_id);
        //pthread_detach(interpreter_list->pthread_id);
        interpreter_list = interpreter_list->next;
    }

    /* start consumer threads */
    while(consumer_list != NULL && consumer_list->consume_event != NULL) {
        rc = pthread_create(&consumer_list->pthread_id, &ta, 
                consumer_thread, (void*)consumer_list);
        if (rc != 0) {
            return -1;
        }
        msg(LOG_DEBUG, "Started consumer thread(0x%08x).",
                consumer_list->pthread_id);
        //pthread_detach(consumer_list->pthread_id);
        consumer_list = consumer_list->next;
    }

    return 0;
}


/* cleanup_threads:
**
** Suspends all threads and dump all event data here.
** No need to process(intereprete) raw event data, just dump all event
** buffer(raw and interpreted) to disk.
*/
void cleanup_plugins(void)
{
    receiver_list_t* receiver_list;
    interpreter_list_t* interpreter_list;
    consumer_list_t* consumer_list;

    /* Cancel receiver threads */
    receiver_list = &plugin_api_list.receiver_list;

    msg(LOG_INFO, "Canceling all receiver threads...");
    do {
        pthread_cancel(receiver_list->pthread_id);
        pthread_testcancel();
        while (receiver_list->error_handler.done != 
                AUDITD_PLUGIN_THREAD_FINISHED) {
            usleep(10000);
        }
        receiver_list = receiver_list->next;
    } while(receiver_list != NULL);
    msg(LOG_INFO, "All receiver threads are killed.");

    /* Cancel interpreter threads */
    interpreter_list = &plugin_api_list.interpreter_list;

    msg(LOG_INFO, "Canceling all interpreter threads...");
    do {
        pthread_cancel(interpreter_list->pthread_id);
        pthread_testcancel();
        while (interpreter_list->error_handler.done != 
                AUDITD_PLUGIN_THREAD_FINISHED) {
            usleep(10000);
        }
        interpreter_list = interpreter_list->next;
    } while(interpreter_list != NULL);
    msg(LOG_INFO, "All interpreter threads are killed.");

    /* Cancel consumer threads */
    consumer_list = &plugin_api_list.consumer_list;

    msg(LOG_INFO, "Canceling all consumer threads...");
    do {
        pthread_cancel(consumer_list->pthread_id);
        pthread_testcancel();
        while (consumer_list->error_handler.done != 
                AUDITD_PLUGIN_THREAD_FINISHED) {
            usleep(10000);
        }
        consumer_list = consumer_list->next;
    } while(consumer_list != NULL);
    msg(LOG_INFO, "All consumer threads are killed.");
}

