/* auditd-plugin.h --
 * Copyright 2005 Linuon Inc. Tokyo Japan
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors:
 *   Junji Kanemaru <junji.kanemaru@linuon.com>
 * 
 */

#ifndef __AUDITD_PLUGIN_H__
#define __AUDITD_PLUGIN_H__

#include <unistd.h>
#include <auditd.h>

#ifdef __cplusplus
extern "C" {
#endif

/* Symbol name of audit plugin. Must be C entry, not C++. */
#define AUDITD_PLUGIN_ENTRY "_auditd_plugin_entry"

/* Plugin type. For example 'DFLT' for default plugin, 'SCAL' for syscall
** aduiting, etc. etc... */
typedef unsigned long auditd_plugin_type_t;

/* Macro to store 4 byte length string into 32bit(=4 bytes) unsigned long type.
** i.e. "1234" -> 0x31323334 */
#include <ctype.h>
#define STR_TO_PLUGIN_TYPE(s) \
    ((auditd_plugin_type_t) (((unsigned long)toupper((s)[0]) & 0xff) << 24) | \
                            (((unsigned long)toupper((s)[1]) & 0xff) << 16) | \
                            (((unsigned long)toupper((s)[2]) & 0xff) <<  8) | \
                            (((unsigned long)toupper((s)[3]) & 0xff) <<  0))

/* Macro for fatal errors */
#define AUDITD_PLUGIN_ASSERT(cond) \
    do { \
        if(!(cond)) { (*env->auditd_fatal_error)(__FILE__, __LINE__, \
            "auditd plugin assertion failed", ERROR_ASSERTION_FAILED, 0); \
            goto fatal_error; \
        } \
    } while(0)

#define AUDITD_PLUGIN_FATAL_ERROR(code, msg) \
    do { \
        (*env->auditd_fatal_error)(__FILE__, __LINE__, (msg), (code), 0); \
        goto fatal_error; \
    } while(0)

/* Plugin private storage */
typedef struct _plugin_private_storage_t {
    unsigned long stat;
    void* buff;
} plugin_private_storage_t;

/* Op code for plugin_private_storage_t */
#define PPS_AVAILABLE    0
#define PPS_ALLOCATED    1
#define PPS_DELETE       2

/* get_type:
** Returns type which the plugin supports. */
typedef auditd_plugin_type_t (*get_type_method_t)(void);

/* There are three types in auditd plugins. One is receiver, one is interpreter,
** and consumer. The receiver receives(reads) events and interpreter interprets
** input audit message and returns formatted string. The consumer takes audit
** message string and process whatever it wants to do, i.e. write log into
** /var/log/message and/or send out message to other daemon via D-BUS or XMLRPC
** for remote host. */

/* Method interfaces. A single plugin may have all interfaces. 
** unsupported functions will be NULL. 
** Each interface method must be re-entrant. */

/* receive_event:
** Receives events. It depends on plugin where it receives event from.
**
** Paramaters:
**  auditd_plugin_type_t: The event type will be set by receive_event
**  buff: Event data ptr. receive_event allocates the buffer
**        Freeing buff is caller's responsible.
**  size: buff size
**
** Return codes:
**  <0: Error code
**   0: No data received
**  >0: Data received(bytes maybe...) */
typedef int (*receive_event_method_t)(
    auditd_plugin_type_t* event_type, 
    void** buff, 
    size_t* size,
    plugin_private_storage_t* pps);

/* interpret_event:
** interpret_event interpretes event data. If passed event data need not
** to be interpreted then interpret_event returns 0(zero).
**
** Paramaters:
**  event_type: The event type
**  in_buff: Event data
**  in_size: buff size
**  interpreted_type: Interpreted type
**  out_buff: Interpreted data
**  out_size: Size of out_buff
**
** Return codes:
**  <0: Error code
**   0: Not interpreted
**  >0: Interpreted((bytes maybe...) */
typedef int (*interpret_event_method_t)(
    auditd_plugin_type_t event_type,
    void* in_buff, 
    size_t in_size, 
    auditd_plugin_type_t* interpreted_type,
    void** out_buff,
    size_t* out_size,
    plugin_private_storage_t* pps);

/* consume_event:
** consume_event processes event. Consumer has responsible to free allocated
** memory by receiver plugin.
**
** Paramaters:
**  type: Interpreted type
**  buff: Event data
**  size: buff size
**
** Return codes:
**  <0: Error code
**   0: Not consumed
**  >0: Event consumed(bytes maybe...) */
typedef int (*consume_event_method_t)(
    auditd_plugin_type_t type,
    void* buff, 
    size_t size,
    plugin_private_storage_t* pps);


/* on_error:
** For real emergency. We might not be able to use audit log facility.
**
** Paramaters:
**  type: Interpreted type
**  buff: Event data
**  size: buff size
**
** Return codes:
**  <0: Error code
**   0: No error, data dumped. */
/* Plugin API structure */
typedef int (*on_error_method_t)(
    auditd_plugin_type_t type,
    void* buff, 
    size_t size);

typedef struct {
    /* get_type_method_t           get_type; */
    receive_event_method_t      receive_event;
    interpret_event_method_t    interpret_event;
    consume_event_method_t      consume_event;
    on_error_method_t           on_error;
} auditd_plugin_api_t;


/* Plugin entry function. Returns plugin API entries.
** Should support subsequent calls. This takes pointer to environment
** which provides auditd's config info and more. It would provide 
** some method interfaces too. */
typedef const auditd_plugin_api_t* (*plugin_entry)(const auditd_env_t* env);

#ifdef __cplusplus
}
#endif
#endif /* __AUDITD_PLUGIN_H__ */

